/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _OSGRENDERACTION_H_
#define _OSGRENDERACTION_H_
#pragma once

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <vector>
#include <stack>
#include <map>
#include <set>

#include <OSGSystemDef.h>
#include <OSGBaseTypes.h>
#include <OSGRenderActionBase.h>
#include <OSGMatrix.h>
#include <OSGMaterial.h>
#include <OSGStatElemTypes.h>

OSG_BEGIN_NAMESPACE

//---------------------------------------------------------------------------
//  Forward References
//---------------------------------------------------------------------------

class Material;
class DrawTreeNode;
class Geometry;
class State;
class Light;
class LightEnv;
class LightChunk;
class ClipPlane;
class SClipPlaneChunk;

class DrawTreeNodeFactory;

//---------------------------------------------------------------------------
//   Types
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  Class
//---------------------------------------------------------------------------

/*! \brief RenderAction class
 */

class OSG_SYSTEMLIB_DLLMAPPING RenderAction : public RenderActionBase
{
  public:

    typedef struct
    {
        UInt32 first;
        Matrix second;
        //Matrix acc;
    }
    MatrixStore;
    
    typedef std::map <Material   *,      DrawTreeNode *         > MaterialMap;
    typedef std::pair<LightChunk *,      Matrix                 > LightStore;
    typedef std::pair<SClipPlaneChunk *, Matrix                 > ClipPlaneStore;

    //-----------------------------------------------------------------------
    //   constants                                                               
    //-----------------------------------------------------------------------

    static StatElemDesc<StatTimeElem>       statDrawTime;
    static StatElemDesc<StatUInt64Elem>     statNMaterials;
    static StatElemDesc<StatUInt64Elem>     statNMatrices;
    static StatElemDesc<StatUInt64Elem>     statNLights;
    static StatElemDesc<StatUInt64Elem>     statNGeometries;
    static StatElemDesc<StatUInt64Elem>     statNTransGeometries;
    static StatElemDesc<StatIntOnceElem>    statNTextures;
    static StatElemDesc<StatIntOnceElem>    statNTexBytes;
    static StatElemDesc<StatStringElem>     statNOcclusionMode;
    static StatElemDesc<StatUInt64Elem>     statNOcclusionTests;
    static StatElemDesc<StatUInt64Elem>     statNOcclusionCulled;

    static const Int32 OcclusionCullingAccurate;
    static const Int32 OcclusionCullingFast;


    //-----------------------------------------------------------------------
    //   enums                                                               
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   types                                                               
    //-----------------------------------------------------------------------
  
    //-----------------------------------------------------------------------
    //   class functions                                                     
    //-----------------------------------------------------------------------

    static RenderAction *create      (void                    );
    
    static void          setPrototype(RenderAction *pPrototype);
    static RenderAction *getPrototype(void                    );


    static void registerEnterDefault (const FieldContainerType &type, 
                                      const Action::Functor    &func);
    
    static void registerLeaveDefault (const FieldContainerType &type, 
                                      const Action::Functor    &func);

    //-----------------------------------------------------------------------
    //   instance functions                                                  
    //-----------------------------------------------------------------------

    virtual ~RenderAction(void); 

    /*------------------------- your_category -------------------------------*/
    
    virtual Action::ResultE start(void       );
    virtual Action::ResultE stop (ResultE res); 

    /*------------------------- your_operators ------------------------------*/

           void          push_matrix(const Matrix &matrix);
           void          pop_matrix (      void          );

    inline const Matrix top_matrix (      void          );

    /*------------------------- assignment ----------------------------------*/

    void dropGeometry(Geometry  *pGeo);
    void dropFunctor (Material::DrawFunctor &func, Material *mat);

    void dropLight     (Light     *pLight);
    void undropLight   (Light     *pLight);

    void dropLightEnv  (LightEnv  *pLightEnv);
    void undropLightEnv(LightEnv  *pLightEnv);

    void dropClipPlane  (ClipPlane     *pClipPlane);
    void undropClipPlane(ClipPlane     *pClipPlane);

    void setStateSorting(bool s);
    bool getStateSorting(void);

    std::vector<Light *> getActiveLights(void);
    UInt32 getActiveLightsMask(void);
    UInt32 getActiveLightsCount(void);


    inline State *getCurrentState(void);

    /*------------------------- comparison ----------------------------------*/

    void setSortTrans(bool bVal);
    bool getSortTrans(void) const;

    void setZWriteTrans(bool bVal);
    bool getZWriteTrans(void) const;

    void setLocalLights(bool bVal);
    bool getLocalLights(void) const;

    void setOcclusionCulling(bool bVal);
    bool getOcclusionCulling(void) const;

    virtual void setUseMulticastSLI(bool on);
    virtual bool getUseMulticastSLI() const;

    virtual void setOcclusionCullingMode(Int32 mode);
    virtual Int32 getOcclusionCullingMode(void) const;

    void setOcclusionCullingThreshold(UInt32 threshold);
    UInt32 getOcclusionCullingThreshold(void) const;

    void setUseGLFinish(bool s);
    bool getUseGLFinish(void) const;

    void setDepthOnlyPass(bool s);
    bool getDepthOnlyPass(void) const;
    void setEffectsPass(bool s);
    bool getEffectsPass(void) const;

    void setNoDepthPassMatTypes(const std::vector<FieldContainerType *> &matTypes);
    std::vector<FieldContainerType *> getNoDepthPassMatTypes(void);

    void setRenderDepthPassOnly(bool s);
    bool getRenderDepthPassOnly(void) const;

    virtual void setSelectionMode(UInt32 m);
    UInt32 getSelectionMode(void) const;

    void setPolygonMode(const OSGGLenum& m);
    const OSGGLenum getPolygonMode(void) const;

    /*------------------------- comparison ----------------------------------*/

    bool isOccluded(DrawTreeNode *pRoot);
    void deleteOcclusionQueriesPool(void);
    OSGGLuint getOcclusionQuery(void);
    OSGGLuint getOcclusionQuery(NodePtr node);
    void setOcclusionQuery(NodePtr node, OSGGLuint occlusionQuery);
    void resetOcclusionQueryIndex(void);
    void setOcclusionMask(NodePtr node, UInt8 mask);
    bool hasGeometryChild(NodePtr node);

    // test a single node
    bool            isVisible( Node* node );
    
    // visibility levels
    bool  pushVisibility(void);
    void  popVisibility(void);

    void (OSG_APIENTRY* _glGenQueriesARB)(OSGGLsizei, OSGGLuint*);
    void (OSG_APIENTRY* _glDeleteQueriesARB)(OSGGLsizei, OSGGLuint*);
    void (OSG_APIENTRY* _glBeginQueryARB)(OSGGLenum, OSGGLuint);
    void (OSG_APIENTRY* _glEndQueryARB)(OSGGLenum);
    void (OSG_APIENTRY* _glGetQueryObjectuivARB)(OSGGLuint, OSGGLenum, OSGGLuint*);

  protected:

    //-----------------------------------------------------------------------
    //   enums                                                               
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   types                                                               
    //-----------------------------------------------------------------------

    typedef RenderActionBase Inherited;

    //-----------------------------------------------------------------------
    //   class variables                                                     
    //-----------------------------------------------------------------------

    static RenderAction    *_pPrototype;

    static std::vector<Functor> *_vDefaultEnterFunctors;
    static std::vector<Functor> *_vDefaultLeaveFunctors;

    //-----------------------------------------------------------------------
    //   class functions                                                     
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   instance variables                                                  
    //-----------------------------------------------------------------------

    DrawTreeNodeFactory      *_pNodeFactory;

    UInt32                    _uiMatrixId;

    MatrixStore               _currMatrix;
    Matrix                    _camInverse;

    std::vector<MatrixStore>  _vMatrixStack;

    MaterialMap               _mMatMap;
    
    //DrawTreeNode             *_pRoot;
    typedef std::map<Int32, DrawTreeNode *> SortKeyMap;
    SortKeyMap                  _pMatRoots;
    
    typedef std::map<Real32, DrawTreeNode *> TransSortMap;
    typedef std::map<Int32, TransSortMap>    TransSortKeyMap;
    TransSortKeyMap             _pTransMatRoots;

    DrawTreeNode               *_pNoStateSortRoot;
    DrawTreeNode               *_pNoStateSortTransRoot;

    typedef std::map<Real32, DrawTreeNode *> OCMap;
    OCMap                       _ocRoot;

    UInt32                    _uiActiveMatrix;
    State                    *_pActiveState;

    UInt32                    _uiNumMaterialChanges;
    UInt32                    _uiNumMatrixChanges;
    UInt32                    _uiNumLightChanges;
    UInt32                    _uiNumGeometries;
    UInt32                    _uiNumTransGeometries;
    UInt32                    _uiNumOcclusionTests;
    UInt32                    _uiNumOcclusionCulled;

    bool                      _bSortTrans;
    bool                      _bZWriteTrans;
    bool                      _bLocalLights;
    bool                      _bOcclusionCulling;
    Int32                     _occlusionCullingMode;
    UInt32                    _occlusionCullingThreshold;
    UInt32                    _currentOcclusionQueryIndex;
    std::vector<NodePtr>      _occluded_nodes;
    std::set<UInt32>          _hier_occlusions;
    UInt32                    _occ_bb_dl;


    Matrix                    _worldToScreenMatrix;
    bool                      _useGLFinish;
    bool					  _effects_pass;
    bool                      _depth_only_pass;
    std::vector<FieldContainerType *> _noDepthPathMatTypes;
    bool                      _render_depth_pass_only;
    UInt32                    _selection_mode;
    OSGGLenum                 _polygon_mode;




    UInt32                    _activeLightsState;
    UInt32                    _activeLightsCount;
    UInt32                    _activeLightsMask;


    UInt32                      _clipPlanesState;
    UInt32                      _activeClipPlanesState;
    UInt32                      _activeClipPlanesCount;
    UInt32                      _activeClipPlanesMask;

    std::vector<std::vector<UInt32> > _clipPlanesTable;
    std::vector<UInt32>               _clipPlanesPath;

    bool                      _stateSorting;

    std::vector<FrustumVolume::PlaneSet>  _visibilityStack;

    OSGGLuint _occlusionQuery;
    std::vector<OSGGLuint> _occlusionQueriesPool;
    std::map<UInt32, OSGGLuint> _occlusionQueries;

    Int32 _shlChunkId;


//    Time                 _tMatSlot

    //-----------------------------------------------------------------------
    //   instance functions                                                  
    //-----------------------------------------------------------------------

    // prohibit default functions (move to 'public' if you need one)

    RenderAction(void);
    RenderAction(const RenderAction &source);

    void operator =(const RenderAction &source);

    // access default functors

    virtual std::vector<Functor> *getDefaultEnterFunctors(void);
    virtual std::vector<Functor> *getDefaultLeaveFunctors(void);

            void dump(DrawTreeNode *pRoot, UInt32 uiIndent);

            void updateShader(State *state);

    virtual void drawDepth(DrawTreeNode *pRoot);
    virtual void draw(DrawTreeNode *pRoot);

    inline  void updateTopMatrix(void);
            void activateLocalLights(DrawTreeNode *pRoot);
            void activateLocalClipPlanes(DrawTreeNode *pRoot);

    void getMaterialStates(Material *mat, std::vector<State *> &states);

    bool isNoDepthPassMat(Material *mat);

  private:

    //-----------------------------------------------------------------------
    //   enums                                                               
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   types                                                               
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   friend classes                                                      
    //-----------------------------------------------------------------------

    friend class ShadowViewport;

    //-----------------------------------------------------------------------
    //   friend functions                                                    
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   class variables                                                     
    //-----------------------------------------------------------------------

    static char cvsid[];
    
    //-----------------------------------------------------------------------
    //   class functions                                                     
    //-----------------------------------------------------------------------

    //-----------------------------------------------------------------------
    //   instance variables                                                  
    //-----------------------------------------------------------------------
    
    //-----------------------------------------------------------------------
    //   instance functions                                                  
    //-----------------------------------------------------------------------
};

//---------------------------------------------------------------------------
//   Exported Types
//---------------------------------------------------------------------------

OSG_END_NAMESPACE

#include "OSGRenderAction.inl"

#endif /* _OSGRENDERACTION_H_ */


