/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifdef OSG_DOC_FILES_IN_MODULE
/*! \file OSGFieldContainer.cpp
    \ingroup GrpSystemFieldContainer
 */
#endif

#include <stdlib.h>
#include <stdio.h>

#include "OSGConfig.h"
#include "OSGFieldContainer.h"

OSG_USING_NAMESPACE

FieldContainerType FieldContainer::_type("FieldContainer");

/*-------------------------------------------------------------------------*/
/*                                Get                                      */

FieldContainerType &FieldContainer::getType(void)
{
    return _type;
}

const FieldContainerType &FieldContainer::getType(void) const
{
    return _type;
}

/*-------------------------------------------------------------------------*/
/*                            Binary Access                                */

SizeT FieldContainer::getBinSize(const BitVector &)
{
    return 0;
}

void FieldContainer::copyToBin(      BinaryDataHandler &,
                               const BitVector         &)
{
}

void FieldContainer::copyFromBin(      BinaryDataHandler &,
                                 const BitVector         &)
{
}

void FieldContainer::initCallbacks()
{
    if(_callbacks == nullptr)
    {
        _callbacks = new std::vector < void* > ;
        _callbacks->push_back(new std::vector< std::weak_ptr<OnDestroyCallback> >);
        _callbacks->push_back(new std::vector< std::weak_ptr<ChangedCallback> >);
    }
}
void FieldContainer::deleteCallbacks()
{
    if (_callbacks != nullptr)
    {
        std::vector< std::weak_ptr<OnDestroyCallback> >* onDestroyCallbacks = reinterpret_cast<std::vector< std::weak_ptr<OnDestroyCallback> >*>((*_callbacks)[0]);
        if(onDestroyCallbacks)
        {
            onDestroyCallbacks->clear();
            delete onDestroyCallbacks;
        }

        std::vector< std::weak_ptr<ChangedCallback> >* onChangeCallbacks = reinterpret_cast<std::vector< std::weak_ptr<ChangedCallback> >*>((*_callbacks)[1]);
        if(onChangeCallbacks)
        {
            onChangeCallbacks->clear();
            delete onChangeCallbacks;
        }

        delete _callbacks;
        _callbacks = nullptr;
    }
}

std::vector< std::weak_ptr<OnDestroyCallback> >* FieldContainer::getOnDestroyCallbacks() const
{
    if (_callbacks != nullptr && _callbacks->size() > 0)
        return reinterpret_cast<std::vector< std::weak_ptr<OnDestroyCallback> >*>((*_callbacks)[0]);

    return nullptr;
}

std::vector< std::weak_ptr<ChangedCallback> >* FieldContainer::getChangeCallbacks() const
{
    if (_callbacks != nullptr && _callbacks->size() > 1)
        return reinterpret_cast<std::vector< std::weak_ptr<ChangedCallback> >*>((*_callbacks)[1]);

    return nullptr;
}


void FieldContainer::registerOnDestroyCallback(std::shared_ptr<OnDestroyCallback> const& callback)
{
    initCallbacks();
    if (getOnDestroyCallbacks() == nullptr)
        return;

    std::vector< std::weak_ptr<OnDestroyCallback> >& onDestroyCallbacks = *getOnDestroyCallbacks();

    bool registered = std::find_if(onDestroyCallbacks.begin(), onDestroyCallbacks.end(),
        [&callback](const std::weak_ptr<OnDestroyCallback>& wp) { return wp.lock() == callback; }) != onDestroyCallbacks.end();

    if (!registered)
    {
        onDestroyCallbacks.push_back(callback);
    }
}

void FieldContainer::unregisterOnDestroyCallback(std::shared_ptr<OnDestroyCallback> const& callback)
{
    if (getOnDestroyCallbacks() == nullptr)
        return;

    std::vector< std::weak_ptr<OnDestroyCallback> >& onDestroyCallbacks = *getOnDestroyCallbacks();

    auto it = std::find_if(onDestroyCallbacks.begin(), onDestroyCallbacks.end(),
        [&callback](const std::weak_ptr<OnDestroyCallback>& wp) { return wp.lock() == callback; });

    if (it != onDestroyCallbacks.end())
    {
        onDestroyCallbacks.erase(it);
    }
}

void FieldContainer::unregisterOnDestroyCallbacks()
{
    if (getOnDestroyCallbacks() != nullptr)
        getOnDestroyCallbacks()->clear();
}

void FieldContainer::registerChangedCallback(std::shared_ptr<ChangedCallback> const& callback)
{
    initCallbacks();
    if (getChangeCallbacks() == nullptr)
        return;

    std::vector< std::weak_ptr<ChangedCallback> >& changedCallbacks = *getChangeCallbacks();

    bool registered = std::find_if(changedCallbacks.begin(), changedCallbacks.end(),
        [&callback](const std::weak_ptr<ChangedCallback>& wp) { return wp.lock() == callback; }) != changedCallbacks.end();

    if (!registered)
    {
        changedCallbacks.push_back(callback);
    }
}

void FieldContainer::unregisterChangedCallback(std::shared_ptr<ChangedCallback> const& callback)
{
    if (getChangeCallbacks() == nullptr)
        return;

    std::vector< std::weak_ptr<ChangedCallback> >& changedCallbacks = *getChangeCallbacks();

    auto it = std::find_if(changedCallbacks.begin(), changedCallbacks.end(),
        [&callback](const std::weak_ptr<ChangedCallback>& wp) { return wp.lock() == callback; });

    if (it != changedCallbacks.end())
    {
        changedCallbacks.erase(it);
    }
}

void FieldContainer::unregisterChangedCallbacks()
{
    if (getChangeCallbacks() != nullptr)
        getChangeCallbacks()->clear();
}


/*-------------------------------------------------------------------------*/
/*                               Changed                                   */

void FieldContainer::changed(BitVector whichField,
                             UInt32    origin    )
{
    // fprintf(stderr, "FC Changed %d %d\n", whichField, fromSync);
    if (_callbacks != nullptr && getChangeCallbacks() != nullptr)
    {
        FieldContainerPtr tmp(this);
        if (tmp != NullFC)
        {
            UInt32 fcid = tmp.getFieldContainerId();
            const std::vector< std::weak_ptr<ChangedCallback> >& changedCallbacks = *getChangeCallbacks();
            for (size_t idx = 0; idx < changedCallbacks.size(); ++idx)
            {
                if (auto callback = changedCallbacks[idx].lock())
                {
                    (*callback)(fcid, whichField, origin);
                }
            }

        }
    }
}

/*-------------------------------------------------------------------------*/
/*                             MT Destruction                              */

void FieldContainer::onDestroy(void)
{
    if (_callbacks != nullptr && getOnDestroyCallbacks() != nullptr)
    {
        FieldContainerPtr tmp(this);
        if (tmp != NullFC)
        {
            UInt32 fcid = tmp.getFieldContainerId();
            const std::vector< std::weak_ptr<OnDestroyCallback> >& onDestroyCallbacks = *getOnDestroyCallbacks();
            for (size_t idx = 0; idx < onDestroyCallbacks.size(); ++idx)
            {
                if (auto callback = onDestroyCallbacks[idx].lock())
                {
                    (*callback)(fcid);
                }
            }
        }
        deleteCallbacks();
    }

}

#if defined(OSG_FIXED_MFIELDSYNC)
void FieldContainer::onDestroyAspect(UInt32 uiId, UInt32 uiAspect)
{
}
#endif

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp[] = "@(#)$Id: $";
    static Char8 cvsid_hpp[] = OSGFIELDCONTAINER_HEADER_CVSID;
    static Char8 cvsid_inl[] = OSGFIELDCONTAINER_INLINE_CVSID;
}
