/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *               Copyright (C) 2000-2002 by the OpenSG Forum                 *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
 \*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
 \*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 \*---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include <sstream>

#include <OSGConfig.h>

#include <OSGGL.h>
#include <OSGGLU.h>
#include <OSGGLEXT.h>
#include <OSGRemoteAspect.h>
#include <OSGCamera.h>
#include <OSGViewport.h>
#include <OSGStereoCameraDecorator.h>
#include <OSGRenderAction.h>
#include <OSGCRC32.h>
#include <OSGFileSystem.h>
#include <OSGCrypt.h>

#include <OSGShaderParameter.h>
#include <OSGShaderParameterBool.h>
#include <OSGShaderParameterInt.h>
#include <OSGShaderParameterUInt.h>
#include <OSGShaderParameterReal.h>
#include <OSGShaderParameterVec2f.h>
#include <OSGShaderParameterVec3f.h>
#include <OSGShaderParameterVec4f.h>
#include <OSGShaderParameterMatrix.h>
#include <OSGShaderParameterHandle.h>

#include <OSGShaderParameterMInt.h>
#include <OSGShaderParameterMReal.h>
#include <OSGShaderParameterMVec2f.h>
#include <OSGShaderParameterMVec3f.h>
#include <OSGShaderParameterMVec4f.h>
#include <OSGShaderParameterMMatrix.h>

#include "OSGSHLChunk.h"

#ifndef BINARY_SHADER_VERSION
#define BINARY_SHADER_VERSION "V1"
#endif

//#define DEBUG_COMPILE_ERROR

#define USE_BINARY true
// disable warnings about missing parameters since there is no chance for us to detect this.
// The shader compiler might optimize the shader based on the uniform variables and completely omit variables that are 
// not used even though they are part of the source code. 
#define OUTPUT_PARAMETER_WARNING 0

OSG_USING_NAMESPACE

/*! \class osg::SHLChunk

*/

/***************************************************************************\
 *                           Class variables                               *
 \***************************************************************************/

 StateChunkClass SHLChunk::_class("SHL");

Int32 SHLChunk::_clusterId = -1;

SHLChunk::parametercbfp SHLChunk::_userParametersCallback = NULL;
SHLChunk::programcbfp SHLChunk::_userCreateProgramCallback = NULL;
SHLChunk::programcbfp SHLChunk::_userDeleteProgramCallback = NULL;

Int32   SHLChunk::s_uniformBlockSize = 16384; // default to the required minimum
bool    SHLChunk::s_hasCommandListSupport = false;
bool    SHLChunk::s_hasStereoViewRenderingSupport = false;

bool    SHLChunk::s_isATI = false;
bool    SHLChunk::s_isIntel = true; // default to intel gpu as the lowest common base
bool    SHLChunk::s_isNvidia = false;

OSG::UInt32         SHLChunk::s_currentProgram = 0;
OSG::Int32          SHLChunk::s_currentGeometryInfoUniform = -1;
// prototypes





/***************************************************************************\
 *                           Class methods                                 *
 \***************************************************************************/

void SHLChunk::initMethod(void)
{
}


/***************************************************************************\
 *                           Instance methods                              *
 \***************************************************************************/

/*-------------------------------------------------------------------------*\
 -  private                                                                 -
 \*-------------------------------------------------------------------------*/

/*----------------------- constructors & destructors ----------------------*/

SHLChunk::SHLChunk(void) :
Inherited(),
_osgParametersCallbacks(),
_oldParameterSize(0),
_fragmentProgram(),
_geometryProgram(),
_vertexProgram(),
_uniforms(),
m_program(0),
m_programChanged(true),
m_programParameterChanged(true),
m_stereoViewRendering(false),
m_singleViewRenderingOnly(false),
_userParameterCallbacks()
{

}

SHLChunk::SHLChunk(const SHLChunk &source) :
Inherited(source),
_osgParametersCallbacks(source._osgParametersCallbacks),
_oldParameterSize(source._oldParameterSize),
m_program(0),
m_programChanged(true),
m_programParameterChanged(true),
m_stereoViewRendering(source.m_stereoViewRendering),
m_singleViewRenderingOnly(source.m_singleViewRenderingOnly),
_userParameterCallbacks(source._userParameterCallbacks)
{
}

SHLChunk::~SHLChunk(void)
{

}

void SHLChunk::onCreate(const SHLChunk *source)
{
    Inherited::onCreate(source);

    // ignore prototypes.
    if (GlobalSystemState == Startup)
        return;

    // we need this for clustering without it handleGL is never called.
    RemoteAspect::addFieldFilter(SHLChunk::getClassType().getId(), SHLChunk::GLIdFieldMask);

}

void SHLChunk::onDestroy(void)
{
    Inherited::onDestroy();

    if (m_program != 0)
    {
        if (osgMakeCurrent())
        {
            if (_userDeleteProgramCallback != NULL)
                _userDeleteProgramCallback(m_program);

            glDeleteProgram(m_program);
            m_program = 0;
            m_programChanged = true;
            m_programParameterChanged = true;
        }
    }
}

const StateChunkClass *SHLChunk::getClass(void) const
{
    return &_class;
}

/*----------------------------- class specific ----------------------------*/

void SHLChunk::parseUniformVariables(void)
{
    _uniforms.clear();
    std::string token;
    std::istringstream is(getVertexProgram() + getFragmentProgram() + getGeometryProgram());
    while (is.good())
    {
        is >> token;
        if (token == "uniform")
        {
            // skip type
            is >> token;
            std::string uniform;
            is >> uniform;
            // remove ;
            if (!uniform.empty())
            {
                if (uniform.substr(uniform.length() - 1) == ";")
                    uniform = uniform.substr(0, uniform.length() - 1);
            }
            _uniforms.insert(uniform);
        }
    }
}

bool SHLChunk::hasUniform(const std::string &name)
{
    return (_uniforms.count(name) > 0);
}

void SHLChunk::changed(BitVector whichField, UInt32 origin)
{

    if ((whichField & VertexProgramFieldMask) ||
        (whichField & FragmentProgramFieldMask) ||
        (whichField & GeometryProgramFieldMask))
    {
        parseUniformVariables();
        m_programChanged = true;
        m_programParameterChanged = true;
    }

    if ((whichField & ParametersFieldMask) ||
        (whichField & ProgramParameterNamesFieldMask) ||
        (whichField & ProgramParameterValuesFieldMask))
    {
        m_programParameterChanged = true;
    }

    Inherited::changed(whichField, origin);
}

void SHLChunk::dump(UInt32,
    const BitVector) const
{
    SLOG << "Dump SHLChunk NI" << std::endl;
}

static void crypt(UInt8 *binary, Int32 size)
{
    static std::vector<UInt8> key;
    if (key.empty())
    {
        key.reserve(16);
        key.push_back('6');
        key.push_back('g');
        key.push_back('a');
        key.push_back('5');
        key.push_back('d');
        key.push_back('8');
        key.push_back('b');
        key.push_back('1');
        key.push_back('e');
        key.push_back('6');
        key.push_back('f');
        key.push_back('f');
        key.push_back('9');
        key.push_back('b');
        key.push_back('a');
        key.push_back('0');
    }

    Crypt c(key);
    for (UInt32 i = 0; i < size; ++i)
        binary[i] = c.code(binary[i]);
}

void SHLChunk::updateProgram()
{
    if (!GLAD_GL_VERSION_4_0)
        return;

    if (m_program != 0 && !m_programChanged && (m_stereoViewRendering == getStereoViewRenderingEnabled()))
        return;
    // build the full shader string
    std::string vertexProg = getShaderVersionString() + getShaderDefinesString(kVertexProgram, m_singleViewRenderingOnly) + getVertexProgram();
    std::string geoProg = getShaderVersionString() + getShaderDefinesString(kGeometryProgram, m_singleViewRenderingOnly) + getGeometryProgram();
    std::string fragmentProg = getShaderVersionString() + getShaderDefinesString(kFragmentProgram, m_singleViewRenderingOnly) + getFragmentProgram();
    
    if (m_program != 0)
    {
        if (_fragmentProgram == fragmentProg &&
            _geometryProgram == geoProg &&
            _vertexProgram == vertexProg)
        {
            m_programChanged = false;
            return; // do nothing, everything is fine
        }

        glUseProgram(0); // remove binding
        if (_userDeleteProgramCallback != NULL)
            _userDeleteProgramCallback(m_program);

        glDeleteProgram(m_program);
        m_program = 0;
        m_programChanged = true;
    }

    m_program = glCreateProgram();

    _fragmentProgram = fragmentProg;
    _geometryProgram = geoProg;
    _vertexProgram = vertexProg;



    CRC32 crc;

    UInt32 vertex_hash = 0;
    UInt32 vShader = 0;
    OSGGLint has_vertex = 0;
    // reload programs
    if (!getVertexProgram().empty())
    {
        crc.fullCRC((const UInt8 *)vertexProg.c_str(), vertexProg.size(), &vertex_hash);
    }

    UInt32 fragment_hash = 0;
    UInt32 fShader = 0;
    OSGGLint has_fragment = 0;
    if (!getFragmentProgram().empty())
    {
        crc.fullCRC((const UInt8 *)fragmentProg.c_str(), fragmentProg.size(), &fragment_hash);
    }

    UInt32 geometry_hash = 0;
    UInt32 gShader = 0;
    OSGGLint has_geometry = 0;
    // reload programs
    if (!getGeometryProgram().empty())
    {
        crc.fullCRC((const UInt8 *)geoProg.c_str(), geoProg.size(), &geometry_hash);
    }
    // create hash filename.
    std::string hash_name;
    if (vertex_hash != 0)
    {
        char hash[20];
        sprintf(hash, "%X", vertex_hash);
        hash_name += hash;
    }
    if (fragment_hash != 0)
    {
        char hash[20];
        sprintf(hash, "%X", fragment_hash);
        hash_name += hash;
    }
    if (geometry_hash != 0)
    {
        char hash[20];
        sprintf(hash, "%X", geometry_hash);
        hash_name += hash;
    }
    bool loaded_binary = false;
    std::string filename;
    UInt8 *binary = NULL;
    Int32 binary_length = 0;
    Int32 binary_format = 0;

    if (USE_BINARY && GLAD_GL_VERSION_4_1) // core since 4.1
    {

        // add gl version + vendor string to get a unique name for different driver versions!
        static std::string glinfo;
        if (glinfo.empty())
        {
            if (glGetString(GL_VERSION) != NULL)
                glinfo += (const char *)glGetString(GL_VENDOR);
            if (glGetString(GL_VERSION) != NULL)
                glinfo += (const char *)glGetString(GL_VERSION);
            glinfo.erase(std::remove(glinfo.begin(), glinfo.end(), ' '), glinfo.end());
        }

        filename = Directory::getTmpPath("OSGGLSLCACHE");
        filename += OSG_DIRSEP_STR;
        filename += hash_name;
        filename += glinfo;
        filename += BINARY_SHADER_VERSION;

        FILE *f = osg::fopen(filename.c_str(), "rb");

        if (f != NULL)
        {
            fseek(f, 0, SEEK_END);
            binary_length = (Int32)ftell(f);
            binary_length -= sizeof(binary_format); // sub sizeof binary format header.
            binary = new UInt8[binary_length];
            fseek(f, 0, SEEK_SET);
            fread(&binary_format, sizeof(binary_format), 1, f);
            fread(binary, binary_length, 1, f);
            // decrypt shader.
            crypt(binary, binary_length);
            fclose(f);
        }
    }

    if (binary != NULL)
    {

        OSGGLenum binaryFormat = (OSGGLenum)binary_format;
        glProgramBinary(m_program, binaryFormat, binary, binary_length);

        delete[] binary;

        OSGGLint success;
        //glGetProgramiv(progId, GL_LINK_STATUS, &success);
        glGetProgramiv(m_program, GL_LINK_STATUS, &success);
        if (success)
        {
            if (_userCreateProgramCallback != NULL)
                _userCreateProgramCallback(m_program);
            loaded_binary = true;
        }
        else
        {
            char *debug = NULL;
            OSGGLint debugLength;
            glGetProgramiv(m_program, GL_INFO_LOG_LENGTH, &debugLength);
            if (debugLength > 0)
            {
                debug = new char[debugLength];
                debug[0] = 0;
                glGetProgramInfoLog(m_program, debugLength, &debugLength, debug);
                if (debug != NULL && debug[0] != 0)
                {
                    FWARNING(("SHLChunk: binary link status: %s\n", debug));
                }
                //else
                //{
                //    FWARNING(("SHLChunk: linked binary shader\n"));
                //}
            }
            delete[] debug;
            // remove binary shader file.
            osg::remove(filename.c_str());
        }
    }

    if (!loaded_binary)
    {
        // reload programs
        if (!getVertexProgram().empty())
        {
            OSGGLenum shader_type = GL_VERTEX_SHADER;
            vShader = glCreateShader(shader_type);
            const char *source[1] = { vertexProg.c_str() };
            glShaderSource(vShader, 1, source, nullptr);

            glCompileShader(vShader);
            glGetShaderiv(vShader, GL_COMPILE_STATUS, &has_vertex);

            if (has_vertex == 0)
            {
                char *debug = NULL;
                OSGGLint debugLength;
                glGetShaderiv(vShader, GL_INFO_LOG_LENGTH, &debugLength);

                if (debugLength > 0)
                {
                    debug = new char[debugLength];
                    debug[0] = 0;
                    glGetShaderInfoLog(vShader, debugLength, &debugLength, debug);
                }

                if (debug != NULL && debug[0] != 0)
                {
                    FFATAL(("Couldn't compile vertex program!\n%s\n", debug));
                }
                else
                {
                    FFATAL(("Couldn't compile vertex program!\n"));
                }
#ifdef DEBUG_COMPILE_ERROR
                std::ofstream outF("VertexProgramError.glsl");
                outF << vertexProg << std::endl;
                outF.close();
#endif
                delete[] debug;
                glDeleteShader(vShader);
            }
        }
        if (!getFragmentProgram().empty())
        {
            OSGGLenum shader_type = GL_FRAGMENT_SHADER;
            fShader = glCreateShader(shader_type);
            const char *source[1] = { fragmentProg.c_str() };
            glShaderSource(fShader, 1, source, nullptr);

            glCompileShader(fShader);
            glGetShaderiv(fShader, GL_COMPILE_STATUS, &has_fragment);

            if (has_fragment == 0)
            {
                char *debug = NULL;
                OSGGLint debugLength;
                glGetShaderiv(fShader, GL_INFO_LOG_LENGTH, &debugLength);

                if (debugLength > 0)
                {
                    debug = new char[debugLength];
                    debug[0] = 0;
                    glGetShaderInfoLog(fShader, debugLength, &debugLength, debug);
                }

                if (debug != NULL && debug[0] != 0)
                {
                    FFATAL(("Couldn't compile fragment program!\n%s\n", debug));
                }
                else
                {
                    FFATAL(("Couldn't compile fragment program!\n"));
                }
#ifdef DEBUG_COMPILE_ERROR
                std::ofstream outF("FragmentProgramError.glsl");
                outF << fragmentProg << std::endl;
                outF.close();
#endif
                delete[] debug;
                glDeleteShader(fShader);
            }
        }
        if (!getGeometryProgram().empty())
        {
            if (GLAD_GL_VERSION_4_0) // core since opengl 4.1
            {
                OSGGLenum shader_type = GL_GEOMETRY_SHADER;

                gShader = glCreateShader(shader_type);
                const char *source[1] = { geoProg.c_str() };
                glShaderSource(gShader, 1, source, nullptr);

                int success = 0;
                glCompileShader(gShader);
                glGetShaderiv(gShader, GL_COMPILE_STATUS, &has_geometry);

                if (has_geometry == 0)
                {
                    char *debug = NULL;
                    OSGGLint debugLength;
                    glGetShaderiv(gShader, GL_INFO_LOG_LENGTH, &debugLength);

                    if (debugLength > 0)
                    {
                        debug = new char[debugLength];
                        debug[0] = 0;
                        glGetShaderInfoLog(gShader, debugLength, &debugLength, debug);
                    }

                    if (debug != NULL && debug[0] != 0)
                    {
                        FFATAL(("Couldn't compile geometry program!\n%s\n", debug));
                    }
                    else
                    {
                        FFATAL(("Couldn't compile geometry program!\n"));
                    }
#ifdef DEBUG_COMPILE_ERROR
                    std::ofstream outF("GeometryProgramError.glsl");
                    outF << geoProg << std::endl;
                    outF.close();
#endif
                    delete[] debug;
                    glDeleteShader(gShader);
                }
            }
            else
            {
                FWARNING(("GL_EXT_geometry_shader4 extension not supported!\n"));
            }
        }
        if (has_vertex || has_fragment || has_geometry)
        {
            if (has_vertex)
            {
                glAttachShader(m_program, vShader);
                // just flagged for deletion
                glDeleteShader(vShader);
            }

            if (has_fragment)
            {
                glAttachShader(m_program, fShader);
                // just flagged for deletion
                glDeleteShader(fShader);
            }

            if (has_geometry)
            {
                glAttachShader(m_program, gShader);
                // just flagged for deletion
                glDeleteShader(gShader);
            }

            if (USE_BINARY && GLAD_GL_VERSION_4_1) // core since 4.1
            {
                const MFGLenum &ppnames = *getMFProgramParameterNames();
                const MFUInt32 &ppvalues = *getMFProgramParameterValues();
                for (UInt32 i = 0; i < ppnames.size(); ++i)
                {
                    if (i < ppvalues.size())
                        glProgramParameteri(m_program, ppnames[i], ppvalues[i]);
                }
                glProgramParameteri(m_program, GL_PROGRAM_BINARY_RETRIEVABLE_HINT, GL_TRUE);
            }
            else
            {
                updateProgramParameters();
            }

            glLinkProgram(m_program);

            OSGGLint success = 0;
            glGetProgramiv(m_program, GL_LINK_STATUS, &success);
            char *debug = NULL;
            OSGGLint debugLength;
            glGetProgramiv(m_program, GL_INFO_LOG_LENGTH, &debugLength);
            if (debugLength > 0)
            {
                debug = new char[debugLength];
                debug[0] = 0;
                glGetProgramInfoLog(m_program, debugLength, &debugLength, debug);
            }

            if (success)
            {
                if (_userCreateProgramCallback != NULL)
                    _userCreateProgramCallback(m_program);

                if (debug != NULL && debug[0] != 0)
                {
                    FWARNING(("SHLChunk: link status: %s\n", debug));
                }
                //else
                //{
                //    FWARNING(("SHLChunk: linked shader\n"));
                //}

                // write binary shader.
                if (USE_BINARY && GLAD_GL_VERSION_4_1)// core since 4.1
                {
                    OSGGLsizei len = 0;
                    // ok for ATI we have to retreive the binary length without this the call to getProgramBinary() doesn't work!
                    glGetProgramiv(m_program, GL_PROGRAM_BINARY_LENGTH, &len);

                    if (len == 0)
                        len = 1048576; // 1 MB fallback
                    UInt8 *binary = new UInt8[len];
                    OSGGLenum binaryFormat = 0;
                    OSGGLsizei rlen = 0;
                    glGetProgramBinary(m_program, len, &rlen, &binaryFormat, binary);
                    if (rlen > 0 && glGetError() == GL_NO_ERROR)
                    {
                        FILE *f = osg::fopen(filename.c_str(), "wb");

                        if (f != NULL)
                        {
                            fwrite(&binaryFormat, sizeof(binaryFormat), 1, f);
                            // crypt shader
                            crypt(binary, rlen);
                            fwrite(binary, rlen, 1, f);
                            fclose(f);
                        }
                    }
                    else
                    {
                        FWARNING(("SHLChunk: couldn't get binary shader program!\n"));
                    }
                    delete[] binary;
                }

                // after linking we can detach the shader objects should save some
                // memory. ok for ATI we have to this after writing the binary shader code,
                // doing it before getProgramBinary fails!
                if (has_vertex)
                    glDetachShader(m_program, vShader);
                if (has_fragment)
                    glDetachShader(m_program, fShader);
                if (has_geometry)
                    glDetachShader(m_program, gShader);
            }
            else
            {
                if (debug != NULL && debug[0] != 0)
                {
                    FFATAL(("Couldn't link vertex and fragment program!\n%s\n", debug));
                }
                else
                {
                    FFATAL(("Couldn't link vertex and fragment program!\n"));
                }
                glUseProgram(0); // remove binding
                glDeleteProgram(m_program);
                m_program = 0;
                m_programChanged = true;
                m_programParameterChanged = true;

            }
            if (debug != NULL)
                delete[] debug;
        }
        else
        {
            glUseProgram(0); // remove binding
            glDeleteProgram(m_program);

            m_program = 0;
            m_programChanged = true;
            m_programParameterChanged = true;
        }
    }
    if (m_program != 0)
    {
        updateProgramParameters();
        // update all parameter locations
        updateParameterLocations(*getMFParameters());
        // update all parameters. Force it since everything has changed
        updateParameters(*getMFParameters(), true, true);

        glUseProgram(m_program);

        // update the standard uniform blocks
        GLuint viewportBlockIndex = glGetUniformBlockIndex(m_program, "OSGViewport");
        if (viewportBlockIndex != GL_INVALID_INDEX)
            glUniformBlockBinding(m_program, viewportBlockIndex, UBO_VIEWPORT);

        GLuint cameraBlockIndex = glGetUniformBlockIndex(m_program, "OSGCamera");
        if (cameraBlockIndex != GL_INVALID_INDEX)
            glUniformBlockBinding(m_program, cameraBlockIndex, UBO_CAMERA);

        GLuint geometryBlockIndex = glGetUniformBlockIndex(m_program, "OSGGeometry");
        if (geometryBlockIndex != GL_INVALID_INDEX)
            glUniformBlockBinding(m_program, geometryBlockIndex, UBO_OBJECT);

        GLuint lightBlockIndex = glGetUniformBlockIndex(m_program, "OSGLightSources");
        if (lightBlockIndex != GL_INVALID_INDEX)
            glUniformBlockBinding(m_program, lightBlockIndex, UBO_LIGHTS);

        if (GLAD_GL_VERSION_4_3 && GLAD_GL_ARB_bindless_texture) // light textures are only supported when bindless is available
        {
            GLuint lightTexturesBlockIndex = glGetUniformBlockIndex(m_program, "OSGLightSourceTextures");
            if (lightTexturesBlockIndex != GL_INVALID_INDEX)
                glUniformBlockBinding(m_program, lightTexturesBlockIndex, UBO_LIGHTTEXTURES);
        }

        GLuint sampleBlockIndex = glGetUniformBlockIndex(m_program, "OSGRandomSamples");
        if (sampleBlockIndex != GL_INVALID_INDEX)
            glUniformBlockBinding(m_program, sampleBlockIndex, UBO_RANDOM);

        GLint ambientOcclussionLocation = glGetUniformLocation(m_program, "OSGAmbientOcclussion");
        if (ambientOcclussionLocation != -1)
            glUniform1i(ambientOcclussionLocation, 13); // always bind the ambient occlusion texture to index 13

        GLint shadowMapLocation = glGetUniformLocation(m_program, "OSGShadowMaps");
        if (shadowMapLocation != -1)
            glUniform1i(shadowMapLocation, 14); // always bind the shadowmaps to index 14

        GLint geometryLightLocation = glGetUniformLocation(m_program, "OSGGeometryLightsTexture");
        if (geometryLightLocation != -1)
            glUniform1i(geometryLightLocation, 15); // always bind the shadowmaps to index 15


        glUseProgram(0);
        m_programChanged = false;
        m_stereoViewRendering = getStereoViewRenderingEnabled();
        
    }
}

void SHLChunk::updateParameterLocation(OSGGLuint program,
    const ShaderParameterPtr &parameter)
{
    if (program == 0)
    {
        FNOTICE(("SHLChunk::updateParameterLocation: program == 0!\n"
            "This frame will probably be rendered wrong!\n"));
        return;
    }

    // as the location won't change after linking we can store them for speedup.
    parameter->setLocation(glGetUniformLocation(program, parameter->getName().c_str()));
}

void SHLChunk::updateParameterLocations( const MFShaderParameterPtr &parameters)
{
    if (m_program == 0 || !m_programChanged && !m_programParameterChanged)
        return;
    //std::cout << "updateParameterLocations " << m_program << " m_programChanged: " << m_programChanged << " m_programParameterChanged: " << m_programParameterChanged << std::endl;
    if (parameters.empty())
        return;

    for (UInt32 i = 0; i < parameters.size(); ++i)
    {
        ShaderParameterPtr parameter = parameters[i];
        // as the location won't change after linking we can store them for speedup.
        parameter->setLocation(glGetUniformLocation(m_program, parameter->getName().c_str()));
    }
}

void SHLChunk::updateParameters(
    const MFShaderParameterPtr &parameters,
    bool useProgram, bool force,
    bool keepProgramActive)
{

    if (!GLAD_GL_VERSION_4_0)
        return;

    if (m_program == 0 || !(m_programChanged || m_programParameterChanged))
    {
        return;
    }
    //std::cout << "updateParameters: " << m_program << " m_programChanged: " << m_programChanged << " m_programParameterChanged: " << m_programParameterChanged << std::endl;

    const MFString &blockNames = *getMFProgramParameterBlockNames();
    if (parameters.empty() && blockNames.empty())
    {
        m_programParameterChanged = false;
        return;
    }
    glUseProgram(m_program);
    // set program parameters.
    const MFUInt32 &blockBindings = *getMFProgramParameterBlockBindings();
    for (UInt32 i = 0; i < blockNames.size(); ++i)
    {
        if (i < blockBindings.size())
        {
            GLuint blockIndex = glGetUniformBlockIndex(m_program, blockNames[i].c_str());
            if (blockIndex != GL_INVALID_INDEX)
                glUniformBlockBinding(m_program, blockIndex, blockBindings[i]);
        }
    }
    for (UInt32 i = 0; i < parameters.size(); ++i)
    {
        ShaderParameterPtr parameter = parameters[i];

        if (!force)
        {
            if (!parameter->hasChanged())
                continue;
            parameter->resetChanged();
        }
        //printf("seeting parameter: '%s'\n", parameter->getName().c_str());

        switch (parameter->getTypeId())
        {
        case ShaderParameter::SHPTypeBool:
        {
            ShaderParameterBoolPtr p = ShaderParameterBoolPtr::dcast(parameter);

            //printf("setting: %s %d\n", p->getName().c_str(), p->getValue());
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform1i(p->getLocation(), OSGGLint(p->getValue()));
#if OUTPUT_PARAMETER_WARNING
            // ok I added my own simple parser to get all uniforms unused uniforms
            // are optimized away from the glsl compiler so we would get here a wrong warning!
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeInt:
        {
            ShaderParameterIntPtr p = ShaderParameterIntPtr::dcast(
                parameter);

            //printf("setting: %s %d\n", p->getName().c_str(), p->getValue());
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform1i(p->getLocation(), p->getValue());
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeUInt:
        {
            ShaderParameterUIntPtr p = ShaderParameterUIntPtr::dcast(
                parameter);

            //printf("setting: %s %d\n", p->getName().c_str(), p->getValue());
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform1ui(p->getLocation(), p->getValue());
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                    p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeReal:
        {
            ShaderParameterRealPtr p = ShaderParameterRealPtr::dcast(parameter);

            //printf("setting: %s %f\n", p->getName().c_str(), p->getValue());
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform1f(p->getLocation(), p->getValue());
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeVec2f:
        {
            ShaderParameterVec2fPtr p = ShaderParameterVec2fPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform2fv(p->getLocation(), 1,
                const_cast<OSGGLfloat *>(
                p->getValue().getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeVec3f:
        {
            ShaderParameterVec3fPtr p = ShaderParameterVec3fPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform3fv(p->getLocation(), 1, const_cast<OSGGLfloat *>(p->getValue().getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeVec4f:
        {
            ShaderParameterVec4fPtr p = ShaderParameterVec4fPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniform4fv(p->getLocation(),
                1,
                const_cast<OSGGLfloat *>(
                p->getValue().getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMatrix:
        {
            ShaderParameterMatrixPtr p = ShaderParameterMatrixPtr::dcast(parameter);
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0)
                glUniformMatrix4fv(p->getLocation(), 1,
                GL_FALSE,
                const_cast<OSGGLfloat *>(
                p->getValue().getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        // arrays
        case ShaderParameter::SHPTypeMInt:
        {
            ShaderParameterMIntPtr p = ShaderParameterMIntPtr::dcast(parameter);

            //printf("setting: %s %d\n", p->getName().c_str(), p->getValue());
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                ; // XXX uniform1iv(p->getLocation(), p->getValue().getSize(), &p->getValue()[0]);
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMReal:
        {
            ShaderParameterMRealPtr p = ShaderParameterMRealPtr::dcast(parameter);
            // get "glUniform1fvARB" function pointer
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                glUniform1fv(p->getLocation(),
                p->getMFValue()->getSize(),
                const_cast<OSGGLfloat *>(&p->getValue(0)));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMVec2f:
        {
            ShaderParameterMVec2fPtr p = ShaderParameterMVec2fPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);

            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                glUniform2fv(p->getLocation(),
                p->getMFValue()->getSize(),
                const_cast<OSGGLfloat *>(
                p->getValue(0).getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMVec3f:
        {
            ShaderParameterMVec3fPtr p = ShaderParameterMVec3fPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                glUniform3fv(p->getLocation(),
                p->getMFValue()->getSize(),
                const_cast<OSGGLfloat *>(
                p->getValue(0).getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMVec4f:
        {
            ShaderParameterMVec4fPtr p = ShaderParameterMVec4fPtr::dcast(parameter);
            // get "glUniform4fvARB" function pointer
            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                glUniform4fv(p->getLocation(),
                p->getMFValue()->getSize(),
                const_cast<OSGGLfloat *>(
                p->getValue(0).getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeMMatrix:
        {
            ShaderParameterMMatrixPtr p = ShaderParameterMMatrixPtr::dcast(parameter);

            if (p->getLocation() == -2)
                updateParameterLocation(m_program, p);
            if (p->getLocation() >= 0 && !p->getMFValue()->empty())
                glUniformMatrix4fv(p->getLocation(),
                p->getMFValue()->getSize(),
                GL_FALSE,
                const_cast<OSGGLfloat *>(
                p->getValue(0).getValues()));
#if OUTPUT_PARAMETER_WARNING
            else if (!hasUniform(p->getName()))
                FWARNING(("Unknown parameter '%s'!\n",
                p->getName().c_str()));
#endif
        }
        break;
        case ShaderParameter::SHPTypeHandle:
        {
            if (GLAD_GL_ARB_bindless_texture)
            {
                ShaderParameterHandlePtr p = ShaderParameterHandlePtr::dcast(parameter);
                if (p->getLocation() == -2)
                    updateParameterLocation(m_program, p);
                if (p->getLocation() >= 0)
                    glUniformHandleui64ARB(p->getLocation(), p->getValue());
#if OUTPUT_PARAMETER_WARNING
                else if (!hasUniform(p->getName()))
                    FWARNING(("Unknown parameter '%s'!\n",
                    p->getName().c_str()));
#endif
            }
        }
        break;
        default:
#if OUTPUT_PARAMETER_WARNING
            FWARNING(("Parameter '%s' has unknown type %d!\n",
                parameter->getName().c_str(),
                parameter->getTypeId()));
#endif
            break;
        }
        parameter->resetChanged();
    }

    if (!keepProgramActive)
    {
        glUseProgram(0);
    }
    m_programParameterChanged = false;
}

void SHLChunk::updateProgramParameters()
{
    if (!GLAD_GL_VERSION_4_0)
        return;

    if (m_program == 0 || !(m_programChanged || m_programParameterChanged))
    {
        return;
    }
   // std::cout << "Update Program Parameters" << m_program  << " m_programChanged: " << m_programChanged  << " m_programParameterChanged: " << m_programParameterChanged  << std::endl;
    // set program parameters.
    const MFGLenum &ppnames = *getMFProgramParameterNames();
    const MFUInt32 &ppvalues = *getMFProgramParameterValues();
    for (UInt32 i = 0; i < ppnames.size(); ++i)
    {
        if (i < ppvalues.size()) {
            glProgramParameteri(m_program, ppnames[i], ppvalues[i]);
        }
    }
}


void SHLChunk::addParameterCallback(const char *name, parametercbfp fp)
{
    if (name == NULL || fp == NULL)
        return;

    //    std::string szName(name);

    osgparametercbfp nullfp = NULL;

    setUniformParameter(name, 0);
    _userParameterCallbacks.insert(
        UserParameterCallbacksMap::value_type(name, std::make_pair(fp, nullfp)));
}

void SHLChunk::addParameterCallback(const char *name, osgparametercbfp fp)
{
    if (name == NULL || fp == NULL)
        return;

    //    std::string szName(name);

    parametercbfp nullfp = NULL;

    setUniformParameter(name, 0);
    _userParameterCallbacks.insert(
        UserParameterCallbacksMap::value_type(name, std::make_pair(nullfp, fp)));
}

void SHLChunk::setParameterCallback(parametercbfp fp)
{
    _userParametersCallback = fp;
}

SHLChunk::parametercbfp SHLChunk::getParameterCallback(void)
{
    return _userParametersCallback;
}

void SHLChunk::setCreateProgramCallback(programcbfp fp)
{
    _userCreateProgramCallback = fp;
}

SHLChunk::programcbfp SHLChunk::getCreateProgramCallback(void)
{
    return _userCreateProgramCallback;
}

void SHLChunk::setDeleteProgramCallback(programcbfp fp)
{
    _userDeleteProgramCallback = fp;
}

SHLChunk::programcbfp SHLChunk::getDeleteProgramCallback(void)
{
    return _userDeleteProgramCallback;
}

void SHLChunk::addProgramParameter(OSGGLenum name, UInt32 value)
{
    editMFProgramParameterNames()->push_back(name);
    editMFProgramParameterValues()->push_back(value);
}

void SHLChunk::subProgramParameter(OSGGLenum name)
{
    MFGLenum &ppnames = *editMFProgramParameterNames();
    MFUInt32 &ppvalues = *editMFProgramParameterValues();

    for (UInt32 i = 0; i < ppnames.size(); ++i)
    {
        if (ppnames[i] == name && i < ppvalues.size())
        {
            ppnames.erase(ppnames.begin() + i);
            ppvalues.erase(ppvalues.begin() + i);
            break;
        }
    }
}

void SHLChunk::setProgramParameter(OSGGLenum name, UInt32 value)
{
    // remove old one.
    subProgramParameter(name);
    // add new one.
    addProgramParameter(name, value);
}

void SHLChunk::addProgramParameterBlockBinding(const char* name, UInt32 value)
{
    editMFProgramParameterBlockNames()->push_back(std::string(name));
    editMFProgramParameterBlockBindings()->push_back(value);
}

void SHLChunk::subProgramParameterBlockBinding(const char* name)
{
    MFString &ppnames = *editMFProgramParameterBlockNames();
    MFUInt32 &ppvalues = *editMFProgramParameterBlockBindings();
    std::string sName = name;
    for (UInt32 i = 0; i < ppnames.size(); ++i)
    {
        if (ppnames[i] == sName && i < ppvalues.size())
        {
            ppnames.erase(ppnames.begin() + i);
            ppvalues.erase(ppvalues.begin() + i);
            break;
        }
    }
}

void SHLChunk::setProgramParameterBlockBinding(const char* name, UInt32 value)
{
    subProgramParameterBlockBinding(name);
    addProgramParameterBlockBinding(name, value);
}

UInt32 SHLChunk::getProgramParameter(OSGGLenum name)
{
    const MFGLenum &ppnames = *getMFProgramParameterNames();
    const MFUInt32 &ppvalues = *getMFProgramParameterValues();

    for (UInt32 i = 0; i < ppnames.size(); ++i)
    {
        if (ppnames[i] == name && i < ppvalues.size())
            return ppvalues[i];
    }
    FWARNING(("SHLChunk::getProgramParameter : Couldn't find program parameter %u!\n",
        name));
    return 0;
}

std::vector<std::pair<OSGGLenum, UInt32> > SHLChunk::getProgramParameters(void)
{
    std::vector<std::pair<OSGGLenum, UInt32> > parameters;

    const MFGLenum &ppnames = *getMFProgramParameterNames();
    const MFUInt32 &ppvalues = *getMFProgramParameterValues();

    for (UInt32 i = 0; i < ppnames.size(); ++i)
    {
        if (i < ppvalues.size())
            parameters.push_back(std::make_pair(ppnames[i], ppvalues[i]));
    }

    return parameters;
}

void SHLChunk::clearProgramParameters(void)
{
    editMFProgramParameterNames()->clear();
    editMFProgramParameterValues()->clear();
}

GLuint SHLChunk::getProgramId()
{
    if (m_program == 0 || m_programChanged || m_programParameterChanged || (m_stereoViewRendering != getStereoViewRenderingEnabled()))
    {
        updateProgram();
        updateProgramParameters();
        updateParameters(*getMFParameters(), true, true);
    }
    return m_program;
}

OSGGLint SHLChunk::getUniformLocation(const char* name)
{
    if (m_program == 0)
        return -1;
    return glGetUniformLocation(m_program, name);
}

OSGGLint SHLChunk::getUniformBlockIndex(const char* name)
{
    if (m_program == 0)
        return -1;
    return glGetUniformBlockIndex(m_program, name);
}


void SHLChunk::setClusterId(Int32 id)
{
    _clusterId = id;
}

void SHLChunk::setUniformBlockSize(Int32 blockSize)
{
    s_uniformBlockSize = blockSize;
}

void SHLChunk::setCommandListSupport(bool on)
{
    s_hasCommandListSupport = on;
}

void SHLChunk::setStereoViewRenderingEnabled(bool on)
{
    s_hasStereoViewRenderingSupport = on;
}

void SHLChunk::setGPUType(bool isIntel, bool isATI, bool isNvidia)
{
    s_isATI = isATI;
    s_isIntel = isIntel;
    s_isNvidia = isNvidia;
}


void SHLChunk::updateClusterId(const ShaderParameterPtr &parameter,
    DrawActionBase *action, OSGGLuint program)
{
    if (parameter->getFlags() & ShaderParameter::SHPFlagUpdate)
        return;

    if (parameter->getLocation() == -2)
        updateParameterLocation(program, parameter);
    if (parameter->getLocation() >= 0)
        glUniform1i(parameter->getLocation(), OSGGLint(_clusterId));
}

/*------------------------------ State ------------------------------------*/

void SHLChunk::requestParameterUpdate()
{
    m_programParameterChanged = true;
}

void SHLChunk::update(DrawActionBase *action)
{

}

void SHLChunk::activate(DrawActionBase *action, UInt32 idx)
{
    if (!GLAD_GL_VERSION_4_0)
        return;

    //std::cout << "updateParameters: " << m_program << " m_programChanged: " << m_programChanged << " m_programParameterChanged: " << m_programParameterChanged << std::endl;
    if (m_program == 0 || m_programChanged || m_programParameterChanged || m_stereoViewRendering != getStereoViewRenderingEnabled())
    {
        updateProgram();
        updateProgramParameters();
        updateParameters(*getMFParameters(), true, true);
    }
    if (getMinShadingSamples() > 0.0f)
        glMinSampleShading(getMinShadingSamples());

    glUseProgram(m_program);

    if (getPointSize())
        enableGLFlag(GL_PROGRAM_POINT_SIZE);

    // set the current active program and geometry Info location
    s_currentProgram = m_program;
}

void SHLChunk::deactivate(DrawActionBase *action, UInt32 idx)
{
    if (!GLAD_GL_VERSION_4_0)
        return;

    glUseProgram(0); // since everything is a shader now, there is not really much need to unbind a program since another bind will come in any case
    
    if (getMinShadingSamples() > 0.0f) // reset
    {
        if (s_isNvidia)
            glMinSampleShading(0.0f); // NVIDIA hase some sort of adaptive code giving good quality by only a slight performance drop
        else// if(s_isATI || s_isIntel)
            glMinSampleShading(0.25f); // on AMD 0.25f has shown to give results similar to nvidia
    }

    if (getPointSize())
        disableGLFlag(GL_PROGRAM_POINT_SIZE);

    s_currentProgram = 0;
    s_currentGeometryInfoUniform = -1;
}

/*-------------------------- Comparison -----------------------------------*/

bool SHLChunk::operator < (const StateChunk &other) const
{
    return this < &other;
}

bool SHLChunk::operator == (const StateChunk &other) const
{
    SHLChunk const *tother = dynamic_cast<SHLChunk const*>(&other);

    if (!tother)
        return false;

    if (getVertexProgram() != tother->getVertexProgram() ||
        getFragmentProgram() != tother->getFragmentProgram() ||
        getMFParameters()->size() != tother->getMFParameters()->size())
        return false;

    return true;
}

bool SHLChunk::operator != (const StateChunk &other) const
{
    return !(*this == other);
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderVersionString()
{
    std::string versionString;
    if (GLAD_GL_VERSION_4_3)
    {
        versionString = "#version 430 core\n";
        versionString += "#define GLAD_GL_VERSION_43\n";
      
    }
    else if (GLAD_GL_VERSION_4_0)
    {
        if (GLAD_GL_MESA)
        {
            versionString = "#version 330 core\n"; //return "#version 430 core\n";
            versionString += "#define GLAD_GL_VERSION_33\n";
        }
        else
        {
            versionString = "#version 400 core\n"; //return "#version 430 core\n";
            versionString += "#define GLAD_GL_VERSION_40\n";
        }
    }
    else
    {
        return "";
    }
    return versionString;
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderStructUniformString()
{
    std::string structString =
        "struct OSGLightSource\n"
        "{\n"
        "    mat4  worldMatrix; // the world transform matrix. \n"
        "    mat4  lightMatrix; // the light matrix for shadow mapping. it is basically the world matrix combined with a projection matrix\n"
        "    vec4  diffuseColor; // diffuse color of the light \n"
        "    vec4  glossyColor; // glossy color of the light \n"
        "    vec4  attenuation; // attenuation: x = constant, y = linear, z = quadratic/realistic, w = undefined, used for padding \n"
        "    vec2  spotCone; // the spotlight cone description: x = cos(innerCone), y = cos(outerCone) . \n"
        "    vec2  nearFar; // near/far clipping used for shadowmap rendering \n"
        "    int   lightType; // int specifying the lightsource type. 0 = directional, 1 = spot, 2 = point light. \n"
        "    float shadowIntensity; // shadow intensity scale\n"
        "    float shadowFilterRadius; // shadow map filter radius\n"
        "    int   shadowMapIndex; // index referring to the shadow map\n"
        "    int   shadowMapResolution; // resolution of the shadow map\n"
        "    int   useColorTexture; // use color map\n"
        "    int   useIESProfile; // use ies profile\n"
        "    float specularVisibility; // specular visibility factor, should be either 0 or 1\n"
        "};\n";

    if (GLAD_GL_VERSION_4_3 && GLAD_GL_ARB_bindless_texture)
    {
        structString +=
            "struct OSGLightSourceTexture\n"
            "{\n"
            "    mat4  lightMatrix; // the light projection matrix. \n"
            "    mat4  invWorldMatrix; // the inverse world transform matrix. \n"
            "    vec2  rotate; // texture rotation\n"
            "    vec2  repeat; // texture repeat \n"
            "    vec2  offset; // texture offset \n"
            "    vec2  reserved; // alignment purposes \n"
            "    sampler2D colorTexture; // color texture sampler\n"
            "    sampler2D iesProfile; // ies profile texture sampler\n"
            "};\n";
    }
    return structString;

    
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderViewportUniformString()
{
    std::string uniformString = "// viewport data, binding id 0\n";

    if (GLAD_GL_VERSION_4_3)
        uniformString += "layout(std140, binding = 0) uniform OSGViewport\n";
    else
        uniformString += "layout(std140) uniform OSGViewport\n";

    uniformString += "{\n"
        "    vec4 region; // left, right, bottom, top\n"
        "    vec4 pixelRegion; // left, right, bottom, top\n"
        "    vec2 pixelResolution; // pixelwidth, pixelheight\n"
        "    vec2 aspect;\n"
        "    float superSamplingScale;\n"
        "    float displayLuminance;\n"
        "    int   lensMatchedShading;\n"
        "    int   reserved1;\n"
        "} viewport;\n";
    return uniformString;
}
/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderCameraUniformString()
{
    std::string uniformString = "// camera matrices and other attributes, binding id 1\n";

    if (GLAD_GL_VERSION_4_3)
        uniformString += "layout(std140, binding = 1) uniform OSGCamera\n";
    else
        uniformString += "layout(std140) uniform OSGCamera\n";

    uniformString += "{\n"
        "    mat4 projectionMatrix[2]; // projection matrix for left and right eye\n"
        "    mat4 viewProjectionMatrix[2]; // view projection matrix for left and right eye\n"
        "    mat4 viewMatrix[2]; // view matrix for left and right eye\n"
        "    mat4 invViewMatrix[2]; // inverse view matrix for left and right eye\n"
        "    vec2 nearfar; // near and far plane\n"
        "    int  headlightActive; // use the headlight or not\n"
        "    int  eye; // the eye currently rendered\n"
        "    OSGLightSource  headlight;\n"
        "} camera;\n";
    return uniformString;
}
/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderObjectUniformString()
{
    std::string uniformString = "// geometry matrices and other attributes, binding id 2\n";

    if (GLAD_GL_VERSION_4_3)
        uniformString += "layout(std140, binding = 2) uniform OSGGeometry\n";
    else
        uniformString += "layout(std140) uniform OSGGeometry\n";

    uniformString += "{\n"
        "    mat4 worldMatrix; // object to world matrix\n"
        "    mat4 invWorldMatrix; // inverse object to world matrix\n"
        "    mat4 transInvWorldMatrix; // transposed inverse world matrix aka normal matrix\n"
        "    vec4 clippingPlanes[6]; // clipping planes\n"
        "    ivec4 activeLightIndices[16]; // index buffer, 16 byte aligned but to get it tightly packed it is an ivec4, allowing 64 local lights\n"
        "    int  clippingPlaneCount; // number of clipping planes\n"
        "    int  activeLightCount; // number of active lightsources\n"
        "} object;\n";

    //uniformString += "uniform vec4 OSGGeometryInfo;\n";

    return uniformString;
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderLightUniformString()
{
    std::string uniformString = "// global lightsource buffer,  binding id 3\n";
    if (GLAD_GL_VERSION_4_3)
        uniformString += "layout(std140, binding = 3) uniform OSGLightSources\n";
    else
        uniformString += "layout(std140) uniform OSGLightSources\n";

    uniformString += "{\n"
        "    int             activeGlobalLightCount; // number of active global lightsources\n"
        "    int             activeGeometryLightCount; // number of active geometry lightsources\n"
        "    OSGLightSource  light[NUMLIGHTS]; // lightsources \n"
        "} lights;\n"
        "uniform sampler2DArrayShadow OSGShadowMaps;\n";

    if (GLAD_GL_VERSION_4_3 && GLAD_GL_ARB_bindless_texture)
    {
        uniformString += "layout(std140, binding = 5) uniform OSGLightSourceTextures\n"
            "{\n"
            "    OSGLightSourceTexture textures[NUMLIGHTS];\n"
            "} lightTextures;\n";
    }
    return uniformString;
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderRandomSamplesUniformString()
{
    std::string uniformString = "// global random samples buffer,  binding id 4\n";

    if (GLAD_GL_VERSION_4_3)
        uniformString += "layout(std140, binding = 4) uniform OSGRandomSamples\n";
    else
        uniformString += "layout(std140) uniform OSGRandomSamples\n";

    uniformString += "{\n"
        "    vec4 value[4096]; // random samples \n"
        "} random;\n";
    return uniformString;
}

/*!
* \brief Returns the shader string for the screen space ambient occlussion texture uniform.
* \param
* \return
*/
std::string SHLChunk::getShaderAmbientOcclussionUniformString()
{
    std::string uniformString = "// global screen space ambient occlussion texture\n";

    uniformString += "uniform sampler2D OSGAmbientOcclussion;\n";

    return uniformString;
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderUniformString()
{
    return getShaderStructUniformString() + getShaderViewportUniformString() + getShaderCameraUniformString() +
        getShaderRandomSamplesUniformString() + getShaderObjectUniformString() + getShaderAmbientOcclussionUniformString();
}

std::string SHLChunk::getShaderDefinesString(ProgramType programType, bool singleViewOnly)
{
    std::string defines;

    if (s_hasCommandListSupport)  // do check if commandlist is supported
    {
        defines += "#extension GL_NV_command_list : enable\n"
            "layout(commandBindableNV) uniform;\n";
    }

    if (GLAD_GL_VERSION_4_3 && GLAD_GL_ARB_bindless_texture)
    {
        defines += "#extension GL_ARB_bindless_texture : require\n";
        defines += "#define GLAD_GL_ARB_bindless_texture\n";
        defines += "layout(bindless_sampler) uniform;\n";
    }

    if (GLAD_GL_NV_viewport_array2)
    {
        defines += "#define GLAD_GL_NV_viewport_array2\n";
        defines += "#extension GL_NV_viewport_array2 : enable\n";


    }
    if (programType == kGeometryProgram)
    {
        if (GLAD_GL_NV_geometry_shader_passthrough)
        {
            defines += "#define GLAD_GL_NV_geometry_shader_passthrough\n";
            defines += "#extension GL_NV_geometry_shader_passthrough : require\n";
        }
    }

    if (GLAD_GL_NV_viewport_array2 && s_hasStereoViewRenderingSupport && !singleViewOnly)
    {
        if (programType == kVertexProgram || (programType == kGeometryProgram))
        {
            defines += "#define GLAD_GL_NV_stereo_view_rendering\n";
            defines += "#extension GL_NV_stereo_view_rendering : require\n";
            defines += "layout(secondary_view_offset = 1) out highp int gl_Layer;\n";
        }
        else if(programType == kFragmentProgram)
        {
            defines += "#define GLAD_GL_NV_stereo_view_rendering\n";
            defines += "in highp int gl_Layer;\n";
            defines += "int eye = gl_Layer;\n";
        }
    }
    else
    {
        defines += "const int eye = 0;\n";
    }


    if (s_uniformBlockSize < 65535)
        defines += "#define NUMLIGHTS 64\n";
    else
        defines += "#define NUMLIGHTS 256\n";

    if (s_isATI)
    {
        defines += "#define _ATI_\n";
    }
    else if (s_isIntel)
    {
        defines += "#define INTEL_GPU\n";
    }

#ifdef __APPLE__
    defines += "#define MAC\n";
#endif
    return defines;
}

/*!
* \brief
* \param
* \return
*/
std::string SHLChunk::getShaderHeaderString(ProgramType programType, bool includeVersion, bool includeLights)
{
    std::string shaderHeaderString = getShaderUniformString();
   /*if (includeVersion)
    {
        shaderHeaderString = getShaderVersionString() + getShaderDefinesString(programType) + getShaderUniformString();
    }
    else
        shaderHeaderString = getShaderDefinesString(programType) +  getShaderUniformString();
    */
    if (includeLights)
        shaderHeaderString += getShaderLightUniformString();

    return shaderHeaderString;
}

/*------------------------------------------------------------------------*/
/*                              cvs id's                                  */

#ifdef OSG_SGI_CC
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp[] = "@(#)$Id: OSGSHLChunk.cpp,v 1.68 2008/12/12 11:28:45 a-m-z Exp $";
    static Char8 cvsid_hpp[] = OSGSHLCHUNKBASE_HEADER_CVSID;
    static Char8 cvsid_inl[] = OSGSHLCHUNKBASE_INLINE_CVSID;

    static Char8 cvsid_fields_hpp[] = OSGSHLCHUNKFIELDS_HEADER_CVSID;
}

#ifdef __sgi
#pragma reset woff 1174
#endif
