/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *               Copyright (C) 2000-2002 by the OpenSG Forum                 *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _OSGSHLCHUNK_H_
#define _OSGSHLCHUNK_H_
#pragma once

#include <OSGConfig.h>

#include <string>
#include <vector>
#include <map>
#include <set>
#include <string>

#include "OSGDrawActionBase.h"
#include "OSGWindow.h"

#include <OSGSHLChunkBase.h>

// Uniform Buffer Ids
#ifndef UBO_VIEWPORT
#define UBO_VIEWPORT 0
#endif

#ifndef UBO_CAMERA
#define UBO_CAMERA 1
#endif

#ifndef UBO_OBJECT 
#define UBO_OBJECT 2
#endif

#ifndef UBO_LIGHTS
#define UBO_LIGHTS 3
#endif

#ifndef UBO_RANDOM
#define UBO_RANDOM 4
#endif

#ifndef UBO_LIGHTTEXTURES
#define UBO_LIGHTTEXTURES 5
#endif


OSG_BEGIN_NAMESPACE

/*! \brief SHLChunk class. See \ref
           PageKernelSHLChunk for a description.
*/

class OSG_SYSTEMLIB_DLLMAPPING SHLChunk : public SHLChunkBase
{
  private:

    typedef SHLChunkBase Inherited;

    /*==========================  PUBLIC  =================================*/
  public:
      enum ProgramType
      {
          kVertexProgram = 0,
          kGeometryProgram = 1,
          kFragmentProgram = 2
      };

    /*---------------------------------------------------------------------*/
    /*! \name                 Chunk Class Access                           */
    /*! \{                                                                 */

    virtual const StateChunkClass * getClass         (void) const;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name              Static Chunk Class Access                       */
    /*! \{                                                                 */

    inline static        UInt32            getStaticClassId (void);
    inline static  const StateChunkClass * getStaticClass   (void);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                      Sync                                    */
    /*! \{                                                                 */

    virtual void changed(BitVector  whichField,
                         UInt32     origin    );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                     Output                                   */
    /*! \{                                                                 */

    virtual void dump(      UInt32     uiIndent = 0,
                      const BitVector  bvFlags  = 0) const;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                       Paramerters                            */
    /*! \{                                                                 */

    static void updateParameterLocation(OSGGLuint program,
                                        const ShaderParameterPtr &parameter);

    void updateParameterLocations(const MFShaderParameterPtr &parameters);

    void updateParameters(const MFShaderParameterPtr &parameters,
                          bool useProgram = true, bool force = false,
                          bool keepProgramActive = false);

    void updateProgramParameters();

    void requestParameterUpdate();

    OSGGLuint   getProgramId();

    OSGGLint    getUniformLocation(const char* name);
    OSGGLint    getUniformBlockIndex(const char* name);

    bool getStereoViewRendering() const;
    void setStereoViewRendering(const bool);

    bool getSingleViewRenderingOnly() const;
    void setSingleViewRenderingOnly(const bool);


    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                       State                                  */
    /*! \{                                                                 */

    virtual void update     ( DrawActionBase * action );

    virtual void activate   ( DrawActionBase * action, UInt32 index = 0 );

    virtual void deactivate ( DrawActionBase * action, UInt32 index = 0 );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Comparison                                */
    /*! \{                                                                 */
    virtual bool   operator <  (const StateChunk &other) const;

    virtual bool   operator == (const StateChunk &other) const;
    virtual bool   operator != (const StateChunk &other) const;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Parameter Callbacks                       */
    /*! \{                                                                 */

    typedef OSGGLint (OSG_APIENTRY * OSGGLGETUNIFORMLOCATIONARBPROC)
            (OSGGLuint programObj, const char *name);

    typedef void (*parametercbfp) (OSGGLGETUNIFORMLOCATIONARBPROC getUniformLocation,
                                  DrawActionBase *action, OSGGLuint program);

    typedef void (*osgparametercbfp) (const ShaderParameterPtr &parameter,
                                      DrawActionBase *action, OSGGLuint program);

    void addParameterCallback(const char *name, parametercbfp fp);
    void addParameterCallback(const char *name, osgparametercbfp fp);

    static void setParameterCallback(parametercbfp fp);
    static parametercbfp getParameterCallback(void);

    typedef void (*programcbfp) (OSGGLuint program);

    static void setCreateProgramCallback(programcbfp fp);
    static programcbfp getCreateProgramCallback(void);

    static void setDeleteProgramCallback(programcbfp fp);
    static programcbfp getDeleteProgramCallback(void);

    static std::string getShaderVersionString();
    static std::string getShaderDefinesString(ProgramType programType, bool singleViewOnly);
    static std::string getShaderStructUniformString();
    static std::string getShaderViewportUniformString();
    static std::string getShaderCameraUniformString();
    static std::string getShaderObjectUniformString();
    static std::string getShaderLightUniformString();
    static std::string getShaderRandomSamplesUniformString();
    static std::string getShaderAmbientOcclussionUniformString();
    static std::string getShaderUniformString();
    static std::string getShaderHeaderString(ProgramType programType, bool includeVersion = true, bool includeLights = false);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Program Parameter                         */
    /*! \{                                                                 */

    void addProgramParameter(OSGGLenum name, UInt32 value);
    void subProgramParameter(OSGGLenum name);
    void setProgramParameter(OSGGLenum name, UInt32 value);
    UInt32 getProgramParameter(OSGGLenum name);
    std::vector<std::pair<OSGGLenum, UInt32> > getProgramParameters(void);
    void clearProgramParameters(void);

    void addProgramParameterBlockBinding(const char* name, UInt32 binding);
    void subProgramParameterBlockBinding(const char* name);
    void setProgramParameterBlockBinding(const char* name, UInt32 binding);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Parameter funcs                           */
    /*! \{                                                                 */

 
    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Cluster                                   */
    /*! \{                                                                 */

    static void setClusterId(Int32 id);
    static void setUniformBlockSize(Int32 blockSize);
    static Int32 getUniformBlockSize();
    static void setCommandListSupport(bool on);
    static bool hasCommandListSupport();
    static void setStereoViewRenderingEnabled(bool on);
    static bool getStereoViewRenderingEnabled();
    
    static void setGPUType(bool isIntel, bool isATI, bool isNvidia);
    /*! \}                                                                 */
    /*=========================  PROTECTED  ===============================*/
  protected:

    // Variables should all be in SHLChunkBase.

    void onCreate(const SHLChunk *source = NULL);
    void onDestroy(void);

    /*---------------------------------------------------------------------*/
    /*! \name                  Constructors                                */
    /*! \{                                                                 */

    SHLChunk(void);
    SHLChunk(const SHLChunk &source);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   Destructors                                */
    /*! \{                                                                 */

    virtual ~SHLChunk(void);

    /*! \}                                                                 */

    /*==========================  PRIVATE  ================================*/
  private:

    friend class FieldContainer;
    friend class SHLChunkBase;

    // class. Used for indexing in State
    static StateChunkClass _class;

    /*---------------------------------------------------------------------*/
    /*! \name            OpenGL Extension Handling                         */
    /*! \{                                                                 */

    
    static Int32 _clusterId;

    static bool  _unknown_parameter_warning;

    /*! \}                                                                 */

    static void initMethod(void);

    // prohibit default functions (move to 'public' if you need one)
    void operator =(const SHLChunk &source);

    void parseUniformVariables(void);
    bool hasUniform(const std::string &name);

    void updateProgram();

   
    static void updateClusterId         (const ShaderParameterPtr &parameter,
                                         DrawActionBase *action, OSGGLuint program);
  
    typedef std::vector<
        std::pair<std::pair<parametercbfp,
                            osgparametercbfp>,
                  ShaderParameterPtr          > > OSGParametersCallbacks;
    typedef std::map<std::string, std::pair<parametercbfp, osgparametercbfp> >
        UserParameterCallbacksMap;

    OSGParametersCallbacks     _osgParametersCallbacks;
    UInt32                     _oldParameterSize;

    std::string                 _fragmentProgram;
    std::string                 _geometryProgram;
    std::string                 _vertexProgram;

    std::set<std::string>       _uniforms;

    OSG::UInt32                 m_program;
    bool                        m_programChanged;
    bool                        m_programParameterChanged;
    bool                        m_stereoViewRendering;
    bool                        m_singleViewRenderingOnly;

    static OSG::UInt32          s_currentProgram;
    static OSG::Int32           s_currentGeometryInfoUniform;

    UserParameterCallbacksMap _userParameterCallbacks;
    static parametercbfp      _userParametersCallback;

    static programcbfp        _userCreateProgramCallback;
    static programcbfp        _userDeleteProgramCallback;

    static OSG::Int32          s_uniformBlockSize; // the maximum size of a uniform buffer block
    static bool                s_hasCommandListSupport; 
    static bool                s_hasStereoViewRenderingSupport;

    static bool                s_isIntel;
    static bool                s_isATI;
    static bool                s_isNvidia;
};

typedef SHLChunk *SHLChunkP;

OSG_END_NAMESPACE

#include <OSGSHLChunkBase.inl>
#include <OSGSHLChunk.inl>

#define OSGSHLCHUNK_HEADER_CVSID "@(#)$Id: OSGSHLChunk.h,v 1.38 2008/12/12 11:28:45 a-m-z Exp $"

#endif /* _OSGCGCHUNK_H_ */
